#include <stdio.h>
#include <avs/flow.h>
#include <avs/avs_data.h>
#include <avs/field.h>

extern float bscall();
extern float implied_vol();

fin()
{
	int out_port;		/* temporary to hold the port number */
	int parm;		/* temporary to hold the parm number */

	/* Set the module name and type */
	AVSset_module_name("finance", MODULE_DATA);

	/* There are no input ports for this module */

	/* Create output ports for the result */
	out_port = AVScreate_output_port("Field Output", "field 1D 2-vector uniform float");
	out_port = AVScreate_output_port("Channel Output", "choice");

	/* Add parameter:  */
	parm = AVSadd_parameter("Read Stock Data Browser", "string", 0, 0, 0);
	AVSconnect_widget(parm,"browser");
	/* Add parameter:  */
	parm = AVSadd_parameter("Read Model Parameter Browser", "string", 0, 0, 0);
	AVSconnect_widget(parm,"browser");

	/* Add parameter:  */
	parm = AVSadd_float_parameter("psi", 1.0, 0.0, 10.0);
	parm = AVSadd_float_parameter("rho", 0.5, -2.0, 2.0);
	parm = AVSadd_float_parameter("volatility(B_S)", 1.0, 0.0, 5.0);
	parm = AVSadd_float_parameter("volatility(Am_S)", 1.0, 0.0, 5.0);

	/* Add parameter: an enable/disable toggle for the scope */
	parm = AVSadd_parameter("sleep", "boolean", 1, 0, 1);
	parm = AVSadd_parameter("step", "oneshot", 1, 0, 1);

	/* There is no compute function for this module */
}

#define MAXV 200
#define PERFRAME 50

main(argc,argv)
int	argc;
char	*argv[];
{
   int fin();
   int sleep,step;
   float *v_b, *v_a, *psi, *rho;
   char *StockFileName, *ParameterFileName;
   AVSfield_float *output, *input;
   char *channel0 = "channel 0",*channel1 = "channel 1";
   char *channel = channel0;
   
   int dims[1];
   int i,j, size = PERFRAME;
   float *tmp,temp;
   float drand48();
   
   FILE *sfp, *pfp;
   int flag;
   float sz,szdiv,k,rate,t,tdiv,div,call, vt_a, vt_b;
   long int total;

   /* coroutine initilization */
   AVScorout_init(argc,argv,fin);
   AVScorout_set_sync(0); 
   dims[0] = size;
   channel = channel0;

   while (1) {
     StockFileName = ParameterFileName = NULL;
     total = 0;
     sleep = step = 1;
     input = NULL;
     AVSparameter_visible("Read Stock Data Browser",1);
     AVSparameter_visible("Read Model Parameter Browser",1);
     do {
       AVScorout_wait();  
       AVScorout_input(&StockFileName, &ParameterFileName, &psi, &rho,&v_b, &v_a,&sleep,&step);
       printf("StockFileName=%s ParameterFileName= %s\n",StockFileName,ParameterFileName);
       printf("v_b = %f psi = %f rho = %f sleep = %d\n",*v_b,*psi,*rho,sleep);
     } while (!StockFileName || !ParameterFileName);
     
     printf("StockFileName=%s ParameterFileName= %s\n",StockFileName,ParameterFileName);
     printf("v_b = %f psi = %f rho = %f sleep = %d\n",*v_b,*psi,*rho,sleep);
     
     if ((sfp = fopen(StockFileName, "r")) == NULL) {
       AVSerror("finance: cannot open %s\n",StockFileName);
       continue;
     }
     
     if ((pfp = fopen(ParameterFileName, "r")) == NULL) {
       AVSerror("finance: cannot open %s\n",ParameterFileName);
       continue;
     }
     
     if (fscanf(pfp, "%f %f",psi, rho) == EOF) {
       AVSerror("finance: incorrect parameter file: %s\n",ParameterFileName);
       continue;
     } 
     AVSmodify_float_parameter("psi",AVS_VALUE, *psi, NULL, NULL);
     AVSmodify_float_parameter("rho",AVS_VALUE, *rho, NULL, NULL);     
     AVSparameter_visible("Read Stock Data Browser",0);
     AVSparameter_visible("Read Model Parameter Browser",0);

     /* Allocate space for graph data */
     input = (AVSfield_float *)AVSdata_alloc("field 1D 2-vector uniform float",dims);
     
     i = 0;
     flag = 1;
     /* first output an initial graph */

     while (1) {
       
       if (sleep) AVScorout_wait();
       
       AVScorout_input(&StockFileName, &ParameterFileName, &psi, &rho,&v_b, &v_a,&sleep,&step);
       printf("v_b = %f psi = %f rho = %f sleep = %d\n",*v_b,*psi,*rho,sleep);
       
       for (i=0; i<(size-1); i++) {
	 *I1DV(input,i) = *I1DV(input,(i+1)) ;  /* 1st vector element */
	 tmp = I1DV(input,i) + 1;               /* 2nd vector element */
	 *tmp = *(I1DV(input,i+1) + 1);
       };
       if (get_data(sfp, &sz, &k, &rate, &t, &tdiv, &div, &flag, &vt_b, &vt_a, &call) == 0) break;
       if (flag == 1) { 
	 AVSmodify_float_parameter("volatility(B_S)",AVS_VALUE,vt_b,NULL,NULL);
	 AVSmodify_float_parameter("volatility(Am_S)",AVS_VALUE,vt_a,NULL,NULL);
	 *v_b = vt_b; *v_a = vt_a;
       }
       /*     printf("volatility = %f\n",*v_b);  */
       szdiv = sz - div*exp(-rate*tdiv);
       temp = bscall(szdiv,k,rate, t, *v_b);
       *I1DV(input,(size-1))=temp = call - temp;
       tmp = I1DV(input,i) + 1;   /*      American Stochastic */
       *tmp = temp - 0.5;
       fprintf(stderr,"total = %li model bias = %f\n",++total,temp);
       output = input;
       AVScorout_exec();
       AVScorout_output(output,channel);  
     }
     fprintf(stderr, "!!!!!!!! Done !!!!!!!!\n");
     fclose(sfp); fclose(pfp);
     AVSfield_free(input);

     /* initilize parameter  */
     AVSmodify_parameter("Read Stock Data Browser",AVS_VALUE, NULL, 0, 0);
     AVSmodify_parameter("Read Model Parameter Browser", AVS_VALUE, NULL, 0, 0);
     AVSmodify_float_parameter("psi",AVS_VALUE, 1.0, 0.0, 10.0);
     AVSmodify_float_parameter("rho",AVS_VALUE, 0.5, -2.0, 2.0);
     AVSmodify_float_parameter("volatility(B_S)",AVS_VALUE,1.0,NULL,NULL);
     AVSmodify_float_parameter("volatility(Am_S)",AVS_VALUE,1.0,NULL,NULL);    
     AVSmodify_parameter("sleep",AVS_VALUE, 1, 0, 1);
     AVSmodify_parameter("step", AVS_VALUE, 1, 0, 1);
   }
 }
/*
  float ran(n)
  float n;
  {   
  return(n*drand48());
  }
  */

int get_data(fp, sz, k, rate, t, tdiv, div, flag, v_b, v_a, call)
FILE *fp;
float *sz, *k, *rate, *t, *tdiv, *div, *v_b, *v_a, *call;
int *flag;
{
  float bid,ask,diva[4],tdiva[4];
  char name[4],rstr1[8],rstr2[8];
  int time,vol,tcall,mature;
  long int date,exp;
  
  static float oldrate;
  int strange = 0;
  
  while (fscanf(fp,"%li %s %i%li%i%f%f%f%f%i%i%i %s %s %f%f%f%f%f%f",
		&date,name,&time,&exp,&mature,k,
		&bid,&ask,sz,&vol,&tcall,flag,rstr1,rstr2,
		&tdiva[1],&diva[1],&tdiva[2],&diva[2],&tdiva[3],&diva[3]) != EOF) 
    {
      if (mature == 0.0) continue;
      /* Some of the records have "." in place of interest rates: */
      if (rstr2[0] != '.')
	sscanf(rstr1,"%f",&oldrate);
      
      *rate = oldrate/100.0;     /* Interest rates are given in percent */
      *call = 0.5*(bid + ask);  /* Average bid and ask to get call price */
      *tdiv = tdiva[1]/365.0;
      *t = mature/365.0;
      
      /* Some records have strike + call <= stock, which is impossible. 
	 Ignore those */
      
      if ((tdiva[2]!=0.0)||(diva[2]!=0.0)||((*flag == 1) && (*k + *call <= *sz))
	  ||((*flag==0) && (strange==1))) {
	fprintf(stderr,"Data error\n");
	continue;
      }
      
      if (*flag == 1) {
	fprintf(stderr,"computer volitility ...\n");
	if ((*v_b = implied_vol(*sz,*k,*rate,*t,*call,*div,*tdiv)) <= 0.0)
	  continue;
      }
      return(1);
    };
    return(0);  /* end of file */
}

	  
